import numpy as np
from typing import Dict, List, Tuple
from enum import Enum
from .utils.invert_matrix import invert_matrix

class CriteriaType(Enum):
    Benefit = 1
    Cost = 0

class FuzzyNumber:
    def __init__(self, a1: float, a2: float, a3: float, a4: float):
        self.a1 = a1
        self.a2 = a2
        self.a3 = a3
        self.a4 = a4
    
    def __repr__(self):
        return f"FuzzyNumber({self.a1}, {self.a2}, {self.a3}, {self.a4})"
    
    def distance(self, other: 'FuzzyNumber') -> float:
        return np.sqrt(0.25 * (
            (self.a1 - other.a1)**2 + 
            (self.a2 - other.a2)**2 + 
            (self.a3 - other.a3)**2 + 
            (self.a4 - other.a4)**2
        ))

class FTOPSISClass:
    
    def __init__(self, linguistic_variables_alternatives: Dict, linguistic_variables_weights: Dict, weights: Dict, criteria_type: Dict,
                 alternatives: List[str], criteria: List[str], profile_matrix: Dict, decision_matrix: Dict, profile_mapping: Dict):
        self.linguistic_variables_alternatives = {k: FuzzyNumber(*v) if not isinstance(v, FuzzyNumber) else v 
                                for k, v in linguistic_variables_alternatives.items()}
        self.linguistic_variables_weights = {k: FuzzyNumber(*v) if not isinstance(v, FuzzyNumber) else v 
                                for k, v in linguistic_variables_weights.items()}
        self.weights = weights
        self.criteria_type = criteria_type
        self.alternatives = alternatives
        self.criteria = criteria
        self.profile_matrix = profile_matrix
        self.decision_matrix = decision_matrix
        self.profile_mapping = profile_mapping

    def _normalize_criteria(self, values: List[FuzzyNumber], criterion: str) -> List[FuzzyNumber]:
        if self.criteria_type[criterion] == CriteriaType.Benefit:
            max_val = max([v.a4 for v in values])  # Usamos o maior valor do suporte
            return [FuzzyNumber(v.a1/max_val, v.a2/max_val, v.a3/max_val, v.a4/max_val) for v in values]
        else:
            min_val = min([v.a1 for v in values])  # Usamos o menor valor do suporte
            return [FuzzyNumber(min_val/v.a4, min_val/v.a3, min_val/v.a2, min_val/v.a1) for v in values]

    def normalize_matrix(self) -> Dict:
        inverted_matrix = invert_matrix(self.alternatives, self.criteria, self.decision_matrix)
        fuzzy_matrix = {
            fund: [self.linguistic_variables_alternatives[term] for term in terms]
            for fund, terms in inverted_matrix.items()
        }

        normalized = {fund: [] for fund in self.alternatives}
        for i, criterion in enumerate(self.criteria):
            values = [fuzzy_matrix[fund][i] for fund in self.alternatives]
            normalized_values = self._normalize_criteria(values, criterion)

            weight = self.linguistic_variables_weights[self.weights[criterion][0]]
            for j, fund in enumerate(self.alternatives):
                nv = normalized_values[j]
                weighted = FuzzyNumber(
                    nv.a1 * weight.a1,
                    nv.a2 * weight.a2,
                    nv.a3 * weight.a3,
                    nv.a4 * weight.a4
                )
                normalized[fund].append(weighted)
        
        return normalized

    def calculate_closeness(self) -> Dict:
        normalized = self.normalize_matrix()

        inverted_profile_matrix = invert_matrix(list(self.profile_mapping.values()), list(self.criteria), self.profile_matrix)

        profiles = list(inverted_profile_matrix.keys())
        results = {fund: {} for fund in self.alternatives}
        
        for profile in profiles:
            A_p_pos = []
            for i, criterion in enumerate(self.criteria):
                term = inverted_profile_matrix[profile][i]
                weight_term = self.weights[criterion]
                
                fuzzy_num = self.linguistic_variables_alternatives[term]
                weight = self.linguistic_variables_weights[weight_term[0]]
                
                weighted = FuzzyNumber(
                    fuzzy_num.a1 * weight.a1,
                    fuzzy_num.a2 * weight.a2,
                    fuzzy_num.a3 * weight.a3,
                    fuzzy_num.a4 * weight.a4
                )
                A_p_pos.append(weighted)
            
            profile_distances = {}
            for other_profile in profiles:
                if other_profile == profile:
                    continue
                
                distance = 0
                for i in range(len(self.criteria)):
                    term_p = inverted_profile_matrix[profile][i]
                    term_op = inverted_profile_matrix[other_profile][i]
                    
                    distance += self.linguistic_variables_alternatives[term_p].distance(
                        self.linguistic_variables_alternatives[term_op]
                    )
                profile_distances[other_profile] = distance
            
            farthest_profile = max(profile_distances, key=profile_distances.get)
            
            A_p_neg = []
            for i, criterion in enumerate(self.criteria):
                term = inverted_profile_matrix[farthest_profile][i]
                weight_term = self.weights[criterion]
                
                fuzzy_num = self.linguistic_variables_alternatives[term]
                weight = self.linguistic_variables_weights[weight_term[0]]
                
                weighted = FuzzyNumber(
                    fuzzy_num.a1 * weight.a1,
                    fuzzy_num.a2 * weight.a2,
                    fuzzy_num.a3 * weight.a3,
                    fuzzy_num.a4 * weight.a4
                )
                A_p_neg.append(weighted)
            
            for fund in self.alternatives:
                d_plus = sum(
                    normalized[fund][i].distance(A_p_pos[i]) 
                    for i in range(len(self.criteria))
                )
                
                d_minus = sum(
                    normalized[fund][i].distance(A_p_neg[i]) 
                    for i in range(len(self.criteria))
                )
                
                cc = d_minus / (d_plus + d_minus) if (d_plus + d_minus) != 0 else 0
                results[fund][profile] = cc
        
        return results

    def classify_funds(self) -> Dict:
        closeness = self.calculate_closeness()
        classification = {}
        
        for fund in self.alternatives:
            max_cc = -1
            best_profile = None
            
            for profile, cc in closeness[fund].items():
                if cc > max_cc:
                    max_cc = cc
                    best_profile = profile
            
            classification[fund] = (best_profile, max_cc)
        
        return classification

    def run(self) -> Tuple[Dict, Dict]:
        closeness = self.calculate_closeness()
        classification = self.classify_funds()
        return closeness, classification


import pandas as pd
from typing import Dict, Any

def format_to_json(resultado: pd.DataFrame, classification_col: str = 'Classificação') -> Dict[str, Any]:
    return {
        "results": {
            supplier: {
                "scores": {
                    col: float(resultado.at[supplier, col])
                    for col in resultado.columns 
                    if col != classification_col
                },
                "classification": resultado.at[supplier, classification_col]
            }
            for supplier in resultado.index
        }
    }


from __future__ import annotations
import json
import numpy as np
import pandas as pd
from typing import Tuple, Dict, Any, Union
from .utils.format_output import format_to_json
from .classes.trapezoidal_core import FuzzyNumber, FTOPSISClass, CriteriaType
from .utils.invert_matrix import invert_matrix

class FTOPSISProcessor:
    
    @staticmethod
    def load_json_data(file_path: str) -> Dict[str, Any]:
        try:
            with open(file_path, 'r') as f:
                return json.load(f)
        except FileNotFoundError:
            raise FileNotFoundError(f"Input file {file_path} not found")
        except json.JSONDecodeError:
            raise ValueError(f"Invalid JSON format in {file_path}")

    @staticmethod
    def print_results(result: pd.DataFrame, title: str = "Results") -> None:
        print(f"\n{title}:")
        print(result.to_string(index=True, float_format="%.5f"))

    @staticmethod
    def print_trapezoidal_results(alternatives: list, closeness: dict, 
                                classification: dict, profile_mapping: dict) -> None:
        profiles = list(profile_mapping.values())
        
        elem_width = 10
        profile_width = 15
        
        # Cabeçalho
        header = f"{'alternative':<{elem_width}}"
        for profile in profiles:
            header += f"{profile:<{profile_width}}"
        header += "Classification"
        print(header)
        
        # Linhas de dados
        for alternative in alternatives:
            cc = closeness[alternative]
            best_profile, best_cc = classification[alternative]
            
            row = f"{alternative:<{elem_width}}"
            for profile in profiles:
                row += f"{cc[profile]:<{profile_width}.5f}"
            row += f"{best_profile} ({best_cc:.5f})"
            
            print(row)

    @staticmethod
    def detect_fuzzy_type(data: Dict[str, Any]) -> str:
        if 'linguistic_variables_alternatives' in data:
            first_term = next(iter(data['linguistic_variables_alternatives'].values()))
            if len(first_term) == 4:
                return 'trapezoidal'
        
        if 'linguistic_variables_alternatives' in data:
            first_term = next(iter(data['linguistic_variables_alternatives'].values()))
            if len(first_term) == 3:
                return 'triangular'
        
        raise ValueError("Could not determine fuzzy number type from JSON structure")

def trapezoidal_ftopsis_class(data: Dict[str, Any]) -> Tuple[pd.DataFrame, Dict]:
    criteria_type = {k: CriteriaType[v] for k, v in data['criteria_type'].items()}
    
    ftopsis = FTOPSISClass(
        linguistic_variables_alternatives=data['linguistic_variables_alternatives'],
        linguistic_variables_weights=data['linguistic_variables_weights'],
        weights=data['weights'],
        criteria_type=criteria_type,
        alternatives=data['alternatives'],
        criteria=data['weights'].keys(),
        profile_matrix=data['profile_matrix'],
        decision_matrix=data['decision_matrix'],
        profile_mapping=data['profile_mapping']
    )
    
    closeness, classification = ftopsis.run()
    
    profiles = list(data['profile_mapping'].values())
    result_data = {
        alternative: {
            **{profile: closeness[alternative][profile] for profile in profiles},
            'Classification': classification[alternative][0]
        }
        for alternative in data['alternatives']
    }
    
    result = pd.DataFrame.from_dict(result_data, orient='index')
    
    json_output = {
        alternative: {
            'scores': {profile: float(closeness[alternative][profile]) for profile in profiles},
            'classification': classification[alternative][0],
        }
        for alternative in data['alternatives']
    }

    return json_output

def triangular_ftopsis_class(data: Dict[str, Any]) -> Tuple[pd.DataFrame, Dict]:
    from .classes.triangular_core import CriteriaType, FTOPSISClass as TriFTOPSIS

    linguistic_vars_alt = {k: np.array(v) for k, v in data['linguistic_variables_alternatives'].items()}
    linguistic_vars_weight = {k: np.array(v) for k, v in data['linguistic_variables_weights'].items()}

    decision_matrix = pd.DataFrame({
        k: [linguistic_vars_alt[val] for val in v]
        for k, v in data['decision_matrix'].items()
    })
    
    profile_matrix = pd.DataFrame({
        k: [linguistic_vars_alt[val] for val in v]
        for k, v in data['profile_matrix'].items()
    })

    weights = pd.DataFrame({k: [linguistic_vars_weight[v[0]]] for k, v in data['weights'].items()})
    criteria_type = {k: CriteriaType[v] for k, v in data['criteria_type'].items()}
    profile_mapping = {int(k): v for k, v in data['profile_mapping'].items()}

    norm_matrix = TriFTOPSIS.normalize_matrix(decision_matrix, criteria_type)
    weighted_matrix = TriFTOPSIS.weigh_matrix(norm_matrix, weights)
    final_matrix = TriFTOPSIS.round_weighted_normalized_matrix(weighted_matrix)

    norm_profile = TriFTOPSIS.normalize_matrix(profile_matrix, criteria_type)
    weighted_profile = TriFTOPSIS.weigh_matrix(norm_profile, weights)
    final_profile = TriFTOPSIS.round_weighted_normalized_matrix(weighted_profile)

    pos_sol, neg_sol = TriFTOPSIS.ideal_solution(final_profile, profile_mapping)
    pos_dist, neg_dist = TriFTOPSIS.distance_calculation(final_matrix, pos_sol, neg_sol)
    result = TriFTOPSIS.proximity_coefficient(pos_dist, neg_dist)

    result.index = data['alternatives']
    result['Classificação'] = result.idxmax(axis=1)
    json_output = format_to_json(result)

    return json_output


def main() -> None:
    while True:
        file_path = 'data/json/trapezoidal_input.json'
        
        #try:
        data = FTOPSISProcessor.load_json_data(file_path)
        fuzzy_type = FTOPSISProcessor.detect_fuzzy_type(data)
        
        if fuzzy_type == 'triangular':
            triangular_ftopsis_class(data)
        else:
            trapezoidal_ftopsis_class(data)
        break

        """    
        except FileNotFoundError:
            print(f"Arquivo não encontrado: {file_path}. Por favor, tente novamente.")
        except ValueError as e:
            print(f"Erro no arquivo: {str(e)}. Por favor, verifique o formato e tente novamente.")
        except Exception as e:
            print(f"Erro inesperado: {str(e)}")
            exit(1)
        """

if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        print(f"\nErro: {str(e)}")
        exit(1)


# ftopsis_class_cin/__init__.py

"""
Pacote FTOPSIS-CLASS: Implementação do método de decisão multi critério.

Uso:
    from ftopsis_class_cin import ftopsis_class_cin_decision_support, trapezoidal_ftopsis_class, triangular_ftopsis_class, FtopsisProcessor

Versão:
    0.1.1
"""

__version__ = "0.1.1"

from .decision import FTOPSISProcessor as FtopsisProcessor, trapezoidal_ftopsis_class, triangular_ftopsis_class
from .main import ftopsis_class_decision_support

__all__ = [
    "FtopsisProcessor",
    "trapezoidal_ftopsis_class",
    "triangular_ftopsis_class",
    "ftopsis_class_decision_support"
]



import pandas as pd
import numpy as np
from enum import Enum


class CriteriaType(Enum):
    Cost = 0
    Benefit = 1


class FTOPSISClass:
    @staticmethod
    def normalize_matrix(
        matrix: pd.DataFrame, criteria_type: dict[str, CriteriaType]
    ) -> pd.DataFrame:
        criteria = list(matrix.columns)

        min_value = []
        max_value = []

        fuzzy_decision_matrix = {}
        division_dictionary = {}

        for x in criteria:
            Type = criteria_type[x]
            if Type == CriteriaType.Cost:
                for i in matrix[x]:
                    min_value.append(min(i))
                division_dictionary[x] = min(min_value)

            elif Type == CriteriaType.Benefit:
                for i in matrix[x]:
                    max_value.append(max(i))
                division_dictionary[x] = max(max_value)

        for y in matrix:
            Type = criteria_type[y]

            if Type == CriteriaType.Cost:
                normalized_fuzzy_number = division_dictionary[y] / matrix[y]
                normalized_list = list(
                    map(
                        lambda row: np.nan_to_num(row, nan=0.0).tolist(),
                        normalized_fuzzy_number,
                    )
                )
                fuzzy_decision_matrix[y] = normalized_list
            elif Type == CriteriaType.Benefit:
                normalized_fuzzy_number = np.nan_to_num(
                    matrix[y] / division_dictionary[y]
                )
                normalized_list = list(
                    map(
                        lambda row: np.nan_to_num(row, nan=0.0).tolist(),
                        normalized_fuzzy_number,
                    )
                )
                fuzzy_decision_matrix[y] = normalized_list

        normalized_decision_matrix = pd.DataFrame(fuzzy_decision_matrix)
        return normalized_decision_matrix

    @staticmethod
    def weigh_matrix(
        df_normalized_matrix: pd.DataFrame, df_vector_weights: pd.DataFrame
    ) -> pd.DataFrame:
        matrix_weight_data = {}

        for col in range(len(df_normalized_matrix.columns)):
            for lin in range(len(df_normalized_matrix)):

                name_column = df_normalized_matrix.columns[col]
                name_column_weights = df_vector_weights.columns[col]

                fuzzy_number = df_normalized_matrix.loc[lin, name_column]
                weight = df_vector_weights.loc[0, name_column_weights]

                weighted_number = fuzzy_number * weight
                key = name_column

                if key not in matrix_weight_data:
                    matrix_weight_data[key] = []
                matrix_weight_data[key].append(weighted_number)

        w_n_matrix = pd.DataFrame(matrix_weight_data)
        weighted_normalized_matrix = FTOPSISClass.round_weighted_normalized_matrix(
            w_n_matrix
        )
        return weighted_normalized_matrix

    @staticmethod
    def round_weighted_normalized_matrix(
        df_weighted_normalized_matrix: pd.DataFrame,
    ) -> pd.DataFrame:
        matrix_data = {}

        for i in df_weighted_normalized_matrix:
            array = []
            for j in df_weighted_normalized_matrix[i]:
                number = np.round(j, 4)
                array.append(number)
            matrix_data[i] = array

        normalized_weighted_rounded_matrix = pd.DataFrame(matrix_data)
        return normalized_weighted_rounded_matrix

    @staticmethod
    def ideal_solution(
        df_profile_matrix: pd.DataFrame, profile_mapping: dict[int, str]
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        positive_ideal_solution_data = {}
        negative_ideal_solution_data = {}

        for index, _ in df_profile_matrix.iterrows():
            positive_ideal_solution_data[f"{profile_mapping[index]}"] = (
                df_profile_matrix.iloc[index]
            )

            if index == 0:
                negative_ideal_solution_data[f"{profile_mapping[index]}"] = (
                    df_profile_matrix.iloc[-1]
                )

            elif index != 0 and index != len(df_profile_matrix) - 1:
                negative_ideal_solution_data[f"{profile_mapping[index]}"] = (
                    df_profile_matrix.iloc[-1]
                )

            else:
                negative_ideal_solution_data[f"{profile_mapping[index]}"] = (
                    df_profile_matrix.iloc[0]
                )

        positive_ideal_solution = pd.DataFrame(positive_ideal_solution_data)
        negative_ideal_solution = pd.DataFrame(negative_ideal_solution_data)

        return positive_ideal_solution, negative_ideal_solution

    def distance_calculation(
        matrix: pd.DataFrame,
        df_positive_solution: pd.DataFrame,
        df_negative_solution: pd.DataFrame,
    ) -> tuple[pd.DataFrame, pd.DataFrame]:
        positive_distance_data = {}
        negative_distance_data = {}

        for i in range(len(matrix)):
            alt = list(matrix.iloc[i])

            for j in df_positive_solution:
                positive_solution = list(df_positive_solution[j])
                distance = FTOPSISClass.euclidean_distance(alt, positive_solution)

                if j not in positive_distance_data:
                    positive_distance_data[j] = []
                positive_distance_data[j].append(sum(distance))

            for j in df_negative_solution:
                negative_solution = list(df_negative_solution[j])
                distance = FTOPSISClass.euclidean_distance(alt, negative_solution)

                if j not in negative_distance_data:
                    negative_distance_data[j] = []
                negative_distance_data[j].append(sum(distance))

        positive_distance = pd.DataFrame(positive_distance_data)
        negative_distance = pd.DataFrame(negative_distance_data)

        return positive_distance, negative_distance

    def euclidean_distance(v1: list[float], v2: list[float]) -> list[float]:
        distance = []

        for v1, v2 in zip(v1, v2):
            d = np.sqrt(1 / len(v1) * sum((x - y) ** 2 for x, y in zip(v1, v2)))
            distance.append(d)

        return distance

    def proximity_coefficient(
        positive_distances: pd.DataFrame, negative_distances: pd.DataFrame
    ) -> pd.DataFrame:
        result = {}

        for col in range(len(positive_distances.columns)):
            for lin in range(len(positive_distances)):
                column_name_p = positive_distances.columns[col]
                column_name_n = negative_distances.columns[col]

                a = positive_distances.loc[lin, column_name_p]
                b = negative_distances.loc[lin, column_name_n]

                calculation = np.divide(b, (a + b))

                key = column_name_p

                if key not in result:
                    result[key] = []
                result[key].append(calculation)

        f_result = pd.DataFrame(result)
        return f_result


from typing import Dict, List

def invert_matrix(elements: List[str], criteria: List[str], decision_matrix_by_criteria: Dict[str, List[str]]) -> Dict[str, List[str]]:
    """
    Inverte a matriz de decisão por critério para uma matriz fuzzy (por elemento), validando os dados.

    Args:
        elements: Lista de elementos (ex: ["F1", "F2", ..., "F10"]).
        criteria: Lista de critérios (ex: ["C1", "C2", ..., "C7"]).
        decision_matrix_by_criteria: Dicionário onde cada chave é um critério e o valor é uma lista de termos.

    Returns:
        Um dicionário no formato {"F1": ["VL", "VL", ...], "F2": [...], ...}.

    Raises:
        KeyError: Se algum critério não estiver presente na matriz.
        ValueError: Se o tamanho das listas de termos for inconsistente.
    """
    # Verificar se todos os critérios existem
    missing_criteria = [c for c in criteria if c not in decision_matrix_by_criteria]
    if missing_criteria:
        raise KeyError(f"Critérios faltando na matriz: {missing_criteria}")
    
    # Verificar o tamanho das listas de termos
    expected_length = len(elements)
    for criterion in criteria:
        actual_length = len(decision_matrix_by_criteria[criterion])
        if actual_length != expected_length:
            raise ValueError(f"Critério '{criterion}' tem {actual_length} termos, mas deveria ter {expected_length}.")
    
    # Inverter a matriz
    fuzzy_decision_matrix = {}
    for i, element in enumerate(elements):
        fuzzy_decision_matrix[element] = [
            decision_matrix_by_criteria[criterion][i]
            for criterion in criteria
        ]
    
    return fuzzy_decision_matrix


from __future__ import annotations
import json
from .decision import FTOPSISProcessor, trapezoidal_ftopsis_class, triangular_ftopsis_class
from typing import Union

def ftopsis_class_decision_support(data: Union[dict, str]):
  if type(data) == str: data = json.loads(data)
  fuzzy_type = FTOPSISProcessor.detect_fuzzy_type(data)
        
  if fuzzy_type == 'triangular': return triangular_ftopsis_class(data)
  return trapezoidal_ftopsis_class(data)
